"""

.. currentmodule:: basin3d.core.synthesis

:platform: Unix, Mac
:synopsis: BASIN-3D ``DataSource`` synthesis classes
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top

"""
import logging
from typing import Iterator, List, Optional

from basin3d.core import monitor
from basin3d.core.models import Base, MeasurementTimeseriesTVPObservation, MonitoringFeature
from basin3d.core.plugin import DataSourcePluginAccess, DataSourcePluginPoint
from basin3d.core.schema.enum import MessageLevelEnum, AggregationDurationEnum
from basin3d.core.schema.query import QueryBase, QueryMeasurementTimeseriesTVP, \
    QueryMonitoringFeature, SynthesisMessage, SynthesisResponse
from basin3d.core.translate import translate_query

logger = monitor.get_logger(__name__)


class MonitorMixin(object):
    """
    Adds monitor log functionality to a class for logging synthesis messages
    """
    def log(self,
            message: str, level: Optional[MessageLevelEnum] = None, where: Optional[List] = None) -> Optional[SynthesisMessage]:
        """
        Add a synthesis message to the synthesis response
        :param message: The message
        :param level:  The message level
        :param where:  Where the message is from
        :return: SynthesisMessage
        """
        logger_level = logging.INFO
        synthesis_message = None
        if level:
            logger_level = level == MessageLevelEnum.CRITICAL and logging.CRITICAL or level == MessageLevelEnum.WARN and logging.WARNING \
                 or level == MessageLevelEnum.ERROR and logging.ERROR or logging.INFO
            synthesis_message = SynthesisMessage(msg=message, level=level, where=where)
        logger.log(logger_level, msg=message, extra=where and {"basin3d_where": ".".join(where)} or {})  # type: ignore
        return synthesis_message

    def info(self, message: str, where: Optional[List] = None):
        """
        Add a info level message
        :param where:
        :param message:
        :return: None
        """
        self.log(message, where=where)

    def warn(self, message: str, where: Optional[List] = None) -> Optional[SynthesisMessage]:
        """
        Add a warning level message
        :param where:
        :param message:
        :return: SynthesisMessage
        """
        return self.log(message, MessageLevelEnum.WARN, where)

    def error(self, message: str, where: Optional[List] = None) -> Optional[SynthesisMessage]:
        """
        Add a error level message
        :param where:
        :param message:
        :return: SynthesisMessage
        """
        return self.log(message, MessageLevelEnum.ERROR, where)

    def critical(self, message: str, where: Optional[List] = None) -> Optional[SynthesisMessage]:
        """
        Add a critical level message
        :param where:
        :param message:
        :return:
        """
        return self.log(message, MessageLevelEnum.CRITICAL, where)


class DataSourceModelIterator(MonitorMixin, Iterator):
    """
    BASIN-3D Data Source Model generator
    """

    @property
    def synthesis_response(self) -> SynthesisResponse:
        """Response object for the Synthesis"""
        return self._synthesis_response

    def __init__(self, query: QueryBase, model_access: 'DataSourceModelAccess'):
        """
        Initialize the generator with the query and the model access

        :param query: the unsynthesized query
        :param model_access: Model access
        """

        self._synthesis_response = SynthesisResponse(query=query)  # type: ignore[call-arg]
        self._model_access: 'DataSourceModelAccess' = model_access

        # Filter the plugins, if specified
        if not self._synthesis_response.query.datasource:
            self._plugins = list(self._model_access.plugins.values())
        elif self._synthesis_response.query.datasource:
            self._plugins = [self._model_access.plugins[d] for d in self._synthesis_response.query.datasource if
                             d in self._model_access.plugins.keys()]

        # Internal attributes that contain the iterator state
        self._plugin_index = -1
        self._model_access_iterator = None
        self._next = None
        self._where: Optional[List] = None
        self._where_context_token = None

    def __next__(self) -> Base:
        """
        Return the next item from the iterator. If there are no further items, raise the StopIteration exception.

        """

        while True:
            # Is there an iterator?  Return the next data item
            if self._model_access_iterator:

                try:
                    self._next = next(self._model_access_iterator)
                    if self._next:
                        return self._next
                except StopIteration as se:
                    # ignore sub iterator StopIteration exception
                    # Get any warnings that may have been generated
                    if hasattr(se, "value") and se.value and se.value.args:
                        if isinstance(se.value.args[0], (list, tuple, set)):
                            for m in se.value.args[0]:
                                self.warn(message=m)
                        else:
                            self.warn("Synthesis generated warnings but they are in the wrong format")

            # Setup to get the data from the next data source plugin
            self._plugin_index += 1
            self._model_access_iterator = None

            # Are there any more plugins?
            if self._plugin_index < len(self._plugins):
                self._where = [self._plugins[self._plugin_index].get_datasource().id,
                               self._model_access.synthesis_model.__name__]
                self._where_context_token = monitor.set_ctx_basin3d_where(self._where)
                plugin: DataSourcePluginPoint = self._plugins[self._plugin_index]

                try:
                    # Get the plugin view and determine if it has a list method
                    plugin_views = plugin.get_plugin_access()
                    if self._model_access.synthesis_model in plugin_views and \
                            hasattr(plugin_views[self._model_access.synthesis_model], "list"):

                        # Now translate the query object
                        translated_query_params: QueryBase = self._model_access.synthesize_query(
                            plugin_views[self._model_access.synthesis_model],
                            self._synthesis_response.query)
                        translated_query_params.datasource = [plugin.get_datasource().id]

                        # Get the model access iterator if synthesized query is valid
                        if translated_query_params.is_valid_translated_query:
                            self._model_access_iterator = plugin_views[self._model_access.synthesis_model].list(
                                query=translated_query_params)
                        else:
                            self.warn(f'Translated query for datasource {plugin.get_datasource().id} is not valid.')

                    else:
                        self.warn("Plugin view does not exist")

                except Exception as e:
                    self.error(f"Unexpected Error({e.__class__.__name__}): {str(e)}")

            else:
                # Clear all context
                if self._where_context_token:
                    monitor.basin3d_where.reset(self._where_context_token)
                raise StopIteration

    def log(self, message: str, level: Optional[MessageLevelEnum] = None, where: Optional[List] = None):  # type: ignore[override]
        """
        Add a synthesis message to the synthesis response
        :param message: The message
        :param level:  The message level
        :return: None
        """
        synthesis_message = super().log(message, level, self._where)
        if synthesis_message:
            self._synthesis_response.messages.append(synthesis_message)


class DataSourceModelAccess(MonitorMixin):
    """
    Base class for DataSource model access.
    """

    def __init__(self, plugins, catalog):
        self._plugins = plugins
        self._catalog = catalog

    @property
    def plugins(self):
        return self._plugins

    @property
    def synthesis_model(self):
        raise NotImplementedError

    def synthesize_query(self, plugin_access: DataSourcePluginAccess,
                         query: QueryBase) -> QueryBase:
        """
        Synthesizes query parameters, if necessary

        :param query: The query information to be synthesized
        :param plugin_access: The plugin view to synthesize query params for
        :return: The synthesized query information
        """
        # do nothing, subclasses may override this
        raise NotImplementedError

    @monitor.ctx_synthesis
    def list(self, query: QueryBase) -> DataSourceModelIterator:
        """
        Return the synthesized plugin results

        :param query: The query for this function
        """
        return DataSourceModelIterator(query, self)

    @monitor.ctx_synthesis
    def retrieve(self, query: QueryBase, **kwargs) -> SynthesisResponse:
        """
        Retrieve a single synthesized value

        :param query: The query for this request
        :param kwargs: messages, list of messages to be returned in the SynthesisResponse
        """
        datasource_id: Optional[str] = None
        datasource = None
        messages: List[Optional[SynthesisMessage]] = []

        if 'messages' in kwargs and kwargs.get('messages'):
            msg = kwargs.get('messages')
            if isinstance(msg, List) and isinstance(msg[0], SynthesisMessage):
                messages.extend(msg)
            else:
                messages.append(self.log('Message mis-configured.', MessageLevelEnum.WARN))

        # if the datasource is in the query, utilize it straight up
        if query.datasource:
            datasource_id = query.datasource[0]

        # otherwise, get from a prefixed field.
        # Not tying specifically to id at this point to enable different retrieve approaches from future classes.
        elif query.prefixed_fields:
            value = None
            for prefixed_field in query.prefixed_fields:
                attr_value = getattr(query, prefixed_field)
                if attr_value and isinstance(attr_value, str):
                    value = attr_value
                    break
                elif attr_value and isinstance(attr_value, list):
                    value = attr_value[0]
                    break

            if value and isinstance(value, str):
                datasource_id = value.split("-")[0]

        try:
            plugin = self.plugins[datasource_id]
            datasource = plugin.get_datasource()
        except KeyError:
            pass

        if datasource:

            plugin_views = plugin.get_plugin_access()
            monitor.set_ctx_basin3d_where([plugin.get_datasource().id, self.synthesis_model.__name__])
            if self.synthesis_model in plugin_views and hasattr(plugin_views[self.synthesis_model], 'get'):

                # Now translate the query object
                translated_query_params: QueryBase = self.synthesize_query(plugin_views[self.synthesis_model], query)
                translated_query_params.datasource = [plugin.get_datasource().id]

                # Get the model access iterator if synthesized query is valid
                if translated_query_params.is_valid_translated_query:
                    item: Optional[Base] = plugin_views[self.synthesis_model].get(query=translated_query_params)
                else:
                    logger.warning(f'Translated query for datasource {plugin.get_datasource().id} is not valid.')

                if item:
                    return SynthesisResponse(query=query, data=item, messages=messages)  # type: ignore[call-arg]
            else:
                messages.append(self.log("Plugin view does not exist", MessageLevelEnum.WARN,
                                         [plugin.get_datasource().id, self.synthesis_model.__name__],))

        else:
            messages.append(self.log("DataSource not found for retrieve request", MessageLevelEnum.ERROR))

        return SynthesisResponse(query=query, messages=messages)  # type: ignore[call-arg]


class MonitoringFeatureAccess(DataSourceModelAccess):
    """
    MonitoringFeature: A feature upon which monitoring is made. OGC Timeseries Profile OM_MonitoringFeature.

    **Properties**

    * *id:* string, Unique feature identifier
    * *name:* string, Feature name
    * *description:* string, Description of the feature
    * *feature_type:* sting, FeatureType: REGION, SUBREGION, BASIN, SUBBASIN, WATERSHED, SUBWATERSHED, SITE, PLOT, HORIZONTAL PATH, VERTICAL PATH, POINT
    * *observed_properties:* list of observed properties (variables) made at the feature. Observed properties are configured via the plugins.
    * *related_sampling_feature_complex:* list of related_sampling features. PARENT features are currently supported.
    * *shape:* string, Shape of the feature: POINT, CURVE, SURFACE, SOLID
    * *coordinates:* location of feature in absolute and/or representative datum
    * *description_reference:* string, additional information about the Feature
    * *related_party:* (optional) list of people or organizations responsible for the Feature
    * *utc_offset:* float, Coordinate Universal Time offset in hours (offset in hours), e.g., +9
    * *url:* url, URL with details for the feature

    **Filter** by the following attributes

    * *datasource (optional):* str, a single data source id prefix
    * *id (optional):* str, a single monitoring feature id. Cannot be combined with monitoring_feature
    * *parent_feature (optional)*: list, a list of parent feature ids. Plugin must have this functionality enabled.
    * *monitoring_feature (optional)*: list, a list of monitoring feature ids. Cannot be combined with id, which will take precedence.

    """
    synthesis_model = MonitoringFeature

    def synthesize_query(self, plugin_access: DataSourcePluginAccess, query: QueryMonitoringFeature) -> QueryBase:  # type: ignore[override]
        """
        Synthesizes query parameters, if necessary

        Parameters Synthesized:

        :param query: The query information to be synthesized
        :param plugin_access: The plugin view to synthesize query params for
        :return: The synthesized query information
        """
        return translate_query(plugin_access, query)

    def retrieve(self, query: QueryMonitoringFeature) -> SynthesisResponse:
        """
        Retrieve the specified Monitoring Feature

        :param query: :class:`basin3d.core.schema.query.QueryMonitoringFeature`, id must be specified; monitoring_feature if specified will be removed.
        :return: The synthesized response containing the specified MonitoringFeature if it exists
        """

        # validate that id is specified
        if not query.id:
            return SynthesisResponse(
                query=query,
                messages=[self.log('query.id field is missing and is required for monitoring feature request by id.',
                                   MessageLevelEnum.ERROR)])  # type: ignore[call-arg]

        msg = []

        # remove monitoring_feature specification (i.e., id takes precedence)
        if query.monitoring_feature:
            mf_text = ', '.join(query.monitoring_feature)
            query.monitoring_feature = None
            msg.append(self.log(f'Monitoring Feature query has both id {query.id} and monitoring_feature {mf_text} '
                                f'specified. Removing monitoring_feature and using id.', MessageLevelEnum.WARN))

        # retrieve / get method order should be: MonitoringFeatureAccess, DataSourceModelAccess, <plugin>MonitoringFeatureAccess.get
        synthesis_response: SynthesisResponse = super().retrieve(query=query, messages=msg)
        return synthesis_response


class MeasurementTimeseriesTVPObservationAccess(DataSourceModelAccess):
    """
    MeasurementTimeseriesTVPObservation: Series of measurement (numerical) observations in
    TVP (time value pair) format grouped by time (i.e., a timeseries).

    **Properties**

    * *id:* string, Observation identifier (optional)
    * *type:* enum, MEASUREMENT_TVP_TIMESERIES
    * *observed_property:* url, URL for the observation's observed property
    * *phenomenon_time:* datetime, datetime of the observation, for a timeseries the start and end times can be provided
    * *utc_offset:* float, Coordinate Universal Time offset in hours (offset in hours), e.g., +9
    * *feature_of_interest:* MonitoringFeature obj, feature on which the observation is being made
    * *feature_of_interest_type:* enum (FeatureTypes), feature type of the feature of interest
    * *result:* dictionary of 2 lists: "value" contains TimeValuePair obj and "quality" the corresponding quality assessment per value, observed values and their quality for the observed property being assessed
    * *time_reference_position:* enum, position of timestamp in aggregated_duration (START, MIDDLE, END)
    * *aggregation_duration:* enum, time period represented by observation (YEAR, MONTH, DAY, HOUR, MINUTE, SECOND, NONE)
    * *unit_of_measurement:* string, units in which the observation is reported
    * *statistic:* list, statistical properties of the observation result (MEAN, MIN, MAX, TOTAL)
    * *result_quality:* list, quality assessment values contained in the result (VALIDATED, UNVALIDATED, SUSPECTED, REJECTED, ESTIMATED)

    **Filter** by the following attributes:

    * *monitoring_feature (required):* List of monitoring_features ids
    * *observed_property (required):* List of observed property variable ids
    * *start_date (required):* date YYYY-MM-DD
    * *end_date (optional):* date YYYY-MM-DD
    * *aggregation_duration (default: DAY):* enum (YEAR|MONTH|DAY|HOUR|MINUTE|SECOND|NONE)
    * *statistic (optional):* List of statistic options, enum (INSTANT|MEAN|MIN|MAX|TOTAL)
    * *datasource (optional):* a single data source id prefix (e.g ?datasource=`datasource.id_prefix`)
    * *result_quality (optional):* enum (VALIDATED|UNVALIDATED|SUSPECTED|REJECTED|ESTIMATED)
    * *sampling_medium (optional):* enum (SOLID_PHASE|WATER|GAS|OTHER)

    """
    synthesis_model = MeasurementTimeseriesTVPObservation

    def synthesize_query(self, plugin_access: DataSourcePluginAccess, query: QueryMeasurementTimeseriesTVP) -> QueryBase:  # type: ignore[override]
        """
        Synthesizes query parameters, if necessary

        Parameters Synthesized:
          + monitoring_feature
          + observed_property
          + aggregation_duration (default: DAY)
          + statistic
          + quality_checked

        :param query:
        :param plugin_access: The plugin view to synthesize query params for
        :return: The query parameters
        """

        # only allow instantaneous data (NONE) or daily data (DAY) data
        # NOTE: query at this point is still in BASIN-3D vocab
        if query.aggregation_duration != AggregationDurationEnum.NONE:
            query.aggregation_duration = AggregationDurationEnum.DAY

        return translate_query(plugin_access, query)



"""

.. currentmodule:: basin3d.synthesis

:synopsis: BASIN-3D Synthesis API
:module author: Val Hendrix <vhendrix@lbl.gov>, Danielle Svehla Christianson <dschristianson@lbl.gov>. Catherine Wong <catwong@lbl.gov>


Functions
----------------
* :func:`register` - Register the specified plugins or implicitly register loaded plugins


Classes
--------
* :class:`DataSynthesizer` - Synthesis API

synthesis.DataSynthesizer Functions
-----------------------------------

* :func:`DataSynthesizer.attribute_mappings`- Search for attribute_mappings which describe how the datasource vocabularies are mapped to BASIN-3D vocabularies, including observed properties, statistics, result_quality, etc.
* :func:`DataSynthesizer.measurement_timeseries_tvp_observations`- Search for Measurement Timeseries TVP Observations from specified Monitoring features and observed property variables
* :func:`DataSynthesizer.monitoring_features`- Search for all monitoring features, features by parent monitoring features, features by monitoring feature identifiers, or look for a single monitoring feature by id.
* :func:`DataSynthesizer.observed_properties`- Search for observed properties

----------------------------------
"""
from importlib import import_module
from typing import List, Union, cast

from basin3d.core.catalog import CatalogSqlAlchemy
from basin3d.core.models import DataSource
from basin3d.core.plugin import PluginMount
from basin3d.core.schema.query import QueryMeasurementTimeseriesTVP, QueryMonitoringFeature, SynthesisResponse
from basin3d.core.synthesis import DataSourceModelIterator, MeasurementTimeseriesTVPObservationAccess, \
    MonitoringFeatureAccess, logger


class SynthesisException(Exception):
    """Special Exception for Synthesis module"""
    pass


def register(plugins: List[str] = None):
    """
    Register the specified plugins or implicitly register loaded plugins


    >>> from basin3d import synthesis
    >>> synthesizer = synthesis.register(['basin3d.plugins.usgs.USGSDataSourcePlugin'])
    >>> synthesizer.datasources
    [DataSource(id='USGS', name='USGS', id_prefix='USGS', location='https://waterservices.usgs.gov/nwis/', credentials={})]

    :param plugins: [Optional] plugins to registered
    :return: DataSynthesizer(plugin_dict, catalog)
    """

    if not plugins:
        # Implicit registration of loaded plugins
        plugins = list(PluginMount.plugins.values())

    if not plugins:
        raise SynthesisException("There are no plugins to register")

    plugin_dict = {}
    catalog = CatalogSqlAlchemy()
    for plugin in plugins:
        if isinstance(plugin, str):
            # If this is a string  convert to module and class then load
            class_name_list = plugin.split(".")
            module_name = plugin.replace(".{}".format(class_name_list[-1]), "")
            module = import_module(module_name)
            plugin_class = getattr(module, class_name_list[-1])
        else:
            # This is already a class
            plugin_class = plugin

        # Instantiate the plugin with the new catalog
        plugin = plugin_class(catalog)
        plugin_dict[plugin_class.get_meta().id_prefix] = plugin

        logger.info("Loading Plugin = {}".format(plugin_class.__name__))

    # Instantiate a synthesizer.
    return DataSynthesizer(plugin_dict, catalog)


class DataSynthesizer:
    """
    Synthesis API
    """

    def __init__(self, plugins: dict, catalog: CatalogSqlAlchemy):
        self._plugins = plugins
        self._catalog = catalog
        self._datasources = {}
        for p in self._plugins.values():
            datasource = p.get_datasource()
            self._datasources[datasource.id] = datasource

        self._catalog.initialize(list(self._plugins.values()))
        self._monitoring_feature_access = MonitoringFeatureAccess(plugins, self._catalog)
        self._measurement_timeseries_tvp_observation_access = \
            MeasurementTimeseriesTVPObservationAccess(plugins, self._catalog)

    @property
    def datasources(self) -> List[DataSource]:
        """
        The Datasources loaded in this synthesizer
        :return:
        """
        return list(self._datasources.values())

    def observed_properties(self):
        """

        >>> from basin3d.plugins import usgs
        >>> from basin3d import synthesis
        >>> synthesizer = synthesis.register()
        >>> response = synthesizer.observed_properties()
        >>> for opv in response:
        ...     print(f'{opv.basin3d_vocab} -- {opv.full_name} -- {opv.units}')
        ACT -- Acetate (CH3COO) -- mM
        Br -- Bromide (Br) -- mM
        Cl -- Chloride (Cl) -- mM
        DIN -- Dissolved Inorganic Nitrogen (Nitrate + Nitrite) -- mg/L
        DTN -- Dissolved Total Nitrogen (DTN) -- mg/L
        F -- Fluoride (F) -- mM
        ...

        BASIN-3D observed properties. An observed property defines what is being measured.
        Data source observed property vocabularies are mapped and thus synthesized to the BASIN-3D observed property vocabulary.

        :return: an iterator of :class:`basin3d.core.models.ObservedProperty` objects

        """
        return self._catalog.find_observed_properties()

    def attribute_mappings(self, datasource_id=None, attr_type=None, attr_vocab=None, from_basin3d=False):
        """

        >>> from basin3d.plugins import usgs
        >>> from basin3d import synthesis
        >>> synthesizer = synthesis.register()
        >>> response = synthesizer.attribute_mappings()  # list all attribute mappings registered
        >>> for attr_mapping in response:
        ...     print(f'{attr_mapping.attr_type} | {attr_mapping.basin3d_vocab} -- {attr_mapping.datasource_vocab}')
        OBSERVED_PROPERTY:SAMPLING_MEDIUM | PH:WATER -- 00400
        OBSERVED_PROPERTY:SAMPLING_MEDIUM | RDC:WATER -- 00060
        OBSERVED_PROPERTY:SAMPLING_MEDIUM | WLE:WATER -- 63161
        OBSERVED_PROPERTY:SAMPLING_MEDIUM | WT:WATER -- 00010
        OBSERVED_PROPERTY:SAMPLING_MEDIUM | DO:WATER -- 00300
        ...

        >>> response = synthesizer.attribute_mappings(datasource_id='USGS', attr_type='STATISTIC')
        >>> for attr_mapping in response:
        ...     print(f'{attr_mapping.attr_type} | {attr_mapping.basin3d_vocab} -- {attr_mapping.datasource_vocab}')
        STATISTIC | MAX -- 00001
        STATISTIC | MIN -- 00002
        STATISTIC | MEAN -- 00003
        STATISTIC | TOTAL -- 00006


        >>> response = synthesizer.attribute_mappings(datasource_id='USGS', attr_type='RESULT_QUALITY', attr_vocab=['VALIDATED', 'ESTIMATED'], from_basin3d=True)
        >>> for attr_mapping in response:
        ...     print(f'{attr_mapping.attr_type} | {attr_mapping.basin3d_vocab} -- {attr_mapping.datasource_vocab}, {attr_mapping.datasource_desc}')
        RESULT_QUALITY | VALIDATED -- A, Approved for publication -- Processing and review completed.
        RESULT_QUALITY | ESTIMATED -- E, Value was computed from estimated unit values.
        RESULT_QUALITY | ESTIMATED -- e, Value has been edited or estimated by USGS personnel and is write protected


        Return all the :class:`basin3d.core.models.AttributMapping` registered or those that match the specified fields.

        :param datasource_id: str, The datasource identifier
        :param attr_type: str, The attribute type (e.g., OBSERVED_PROPERTY, STATISTIC, etc)
        :param attr_vocab: str, The attribute vocabulary, either the BASIN-3D vocabulary or the datasource vocabulary
        :param from_basin3d: bool, True = the specified attr_vocab is a BASIN-3D vocabulary, False: the specified attr_vocab is from the datasource

        :return: iterator of :class:`basin3d.core.models.AttributeMapping` objects
        """
        return self._catalog.find_attribute_mappings(datasource_id=datasource_id, attr_type=attr_type, attr_vocab=attr_vocab, from_basin3d=from_basin3d)

    def monitoring_features(self, query: QueryMonitoringFeature = None, **kwargs) -> Union[DataSourceModelIterator, SynthesisResponse]:
        """
        Search for all Monitoring Features, Monitoring Features by parent monitoring features, or Monitoring Feature by id(s).

        To see feature types for a given plugin: **<plugin_module>.<plugin_class>.feature_types**

        **Search for a single monitoring feature by id:**

        >>> from basin3d.plugins import usgs, epa
        >>> from basin3d import synthesis
        >>> synthesizer = synthesis.register()
        >>> response = synthesizer.monitoring_features(id='USGS-0101')
        >>> mf = response.data
        >>> print(f"{mf.id} - {mf.description}")
        USGS-0101 - SUBREGION: St. John

        **Search for all USGS monitoring features:**

        >>> for mf in synthesizer.monitoring_features(datasource='USGS', feature_type='region'): # doctest: +ELLIPSIS
        ...     print(f"{mf.id} - {mf.description}")
        USGS-01 - REGION: New England
        USGS-02 - REGION: Mid Atlantic
        USGS-03 - REGION: South Atlantic-Gulf
        ...

        **Search for USGS points by parent (subbasin) monitoring features:**

        >>> for mf in synthesizer.monitoring_features(feature_type='point',parent_feature=['USGS-17040101']): # doctest: +ELLIPSIS
        ...    print(f"{mf.id} {mf.coordinates and [(p.x, p.y) for p in mf.coordinates.absolute.horizontal_position]}")
        USGS-13010000 [(-110.6647222, 44.1336111)]
        USGS-13010065 [(-110.6675, 44.09888889)]
        USGS-13010450 [(-110.5874305, 43.9038296)]
        ...

        **Search for USGS points by monitoring features identifiers:**

        >>> for mf in synthesizer.monitoring_features(feature_type='point', monitoring_feature=['USGS-13010000', 'USGS-13010450']): # doctest: +ELLIPSIS
        ...    print(f"{mf.id} {mf.coordinates and [(p.x, p.y) for p in mf.coordinates.absolute.horizontal_position]}")
        USGS-13010000 [(-110.6647222, 44.1336111)]
        USGS-13010450 [(-110.5874305, 43.9038296)]

        **Unsupported feature types warning:**

        The code below is an example of what you will see if a registered plugin does not support
        the requested feature type.

        >>> response_itr = synthesizer.monitoring_features(feature_type='horizontal_path')
        >>> for mf in response_itr:
        ...   print(mf)
        ...


        **Output warning messages from the returned iterator**

        This is an example of checking the synthesis response messages in the :class:`basin3d.core.synthesis.DataSourceModelIterator`.

        >>> response_itr.synthesis_response.messages
        [SynthesisMessage(msg='Feature type HORIZONTAL_PATH not supported by USGS.', level='WARN', where=['USGS', 'MonitoringFeature']), SynthesisMessage(msg='Feature type HORIZONTAL_PATH not supported by EPA Water Quality eXchange.', level='WARN', where=['EPA', 'MonitoringFeature'])]

        :param query: (optional) A Monitoring Feature Query :class:`basin3d.core.schema.query.QueryMonitoringFeature` object
        :param kwargs: (optional) Monitoring Feature Query parameters. See Query info below.
        :return: a single :class:`~basin3d.core.schema.query.SynthesisResponse` for a query by id
            or a :class:`~basin3d.core.synthesis.DataSourceModelIterator` for multple :class:`basin3d.core.models.MonitoringFeature` objects.

        .. note::
            **Monitoring Feature Query parameters**
            :class:`basin3d.core.schema.query.QueryMonitoringFeature`

            | All parameters are optional.
            |
            |    * **datasource** A single data source id prefix.
            |    * **feature_type** The :class:`basin3d.core.schema.enum.FeatureTypeEnum` of the desired Monitoring Feature(s). Data Sources may not support all Feature Types.
            |
            | Only one of the following can be specified in a query:
            |
            |    * **id** A single Monitoring Feature ID for the Monitoring Feature desired. Returns a single Synthesis Response object.
            |    * **monitoring_feature** List of Monitoring Feature IDs for the Monitoring Features desired. Returns iterator of MonitoringFeature objects.
            |    * **parent_feature** List of Monitoring Feature IDs for the parent features of the desired Monitoring Features. Returns iterator of MonitoringFeature objects.


        .. note::
            **Monitoring Feature attributes**
            :class:`basin3d.core.models.MonitoringFeature`

            | Attributes values are dependent on data source features.
            |
            |    * **id** Unique feature identifier, prefixed by data source id
            |    * **name** Feature name
            |    * **description** Description of the Monitoring Feature
            |    * **feature_type** :class:`basin3d.core.schema.enum.FeatureTypeEnum` REGION, SUBREGION, BASIN, SUBBASIN, WATERSHED, SUBWATERSHED, SITE, PLOT, HORIZONTAL PATH, VERTICAL PATH, POINT
            |    * **observed_properties** List of observed properties :class:`basin3d.core.models.ObservedProperty` collected at the feature.
            |    * **related_sampling_feature_complex** List of :class:`basin3d.core.models.RelatedSamplingFeature`. PARENT features are currently supported.
            |    * **shape** Shape of the feature: POINT, CURVE, SURFACE, SOLID
            |    * **coordinates** Location of feature in absolute and/or representative datum: :class:`basin3d.core.models.Coordinate`
            |    * **description_reference** Additional information about the feature
            |    * **related_party** List of people or organizations responsible for the feature
            |    * **utc_offset** Coordinate Universal Time offset in hours (offset in hours), e.g., +9
            |    * **datasource** The feature's data source :class:`basin3d.core.models.DataSource`

        """

        if not query:
            query = QueryMonitoringFeature(**kwargs)

        # Search for single or list?
        if query.id:
            #  mypy casts are only used as hints for the type checker,
            #  and they don’t perform a runtime type check.
            return cast(SynthesisResponse, self._monitoring_feature_access.retrieve(query=query))
        else:
            #  mypy casts are only used as hints for the type checker,
            #  and they don’t perform a runtime type check.
            return cast(DataSourceModelIterator,
                        self._monitoring_feature_access.list(query=query))

    def measurement_timeseries_tvp_observations(self, query: QueryMeasurementTimeseriesTVP = None, **kwargs) -> DataSourceModelIterator:
        """
        Search for Measurement Timeseries TVP Observations for the specified query arguments.

        Aggregation Duration for DAY (default) and NONE are both supported.

            **Search with aggregation duration DAY:**

            >>> from basin3d.plugins import usgs
            >>> from basin3d import synthesis
            >>> synthesizer = synthesis.register()
            >>> timeseries = synthesizer.measurement_timeseries_tvp_observations(monitoring_feature=['USGS-09110990'],observed_property=['RDC','WT'],start_date='2019-10-01',end_date='2019-10-30',aggregation_duration='DAY')
            >>> for timeseries in timeseries:
            ...    print(f"{timeseries.feature_of_interest.id} - {timeseries.observed_property.get_basin3d_vocab()}")
            USGS-09110990 - RDC

            **Search with aggregation duration NONE:**

            >>> from basin3d.plugins import usgs
            >>> from basin3d import synthesis
            >>> synthesizer = synthesis.register()
            >>> timeseries = synthesizer.measurement_timeseries_tvp_observations(monitoring_feature=["USGS-09110990", "USGS-09111250"],observed_property=['RDC','WT'],start_date='2020-04-01',end_date='2020-04-30',aggregation_duration='NONE')
            >>> for timeseries in timeseries:
            ...    print(f"{timeseries.feature_of_interest.id} - {timeseries.observed_property.get_basin3d_vocab()}")
            USGS-09110990 - RDC
            USGS-09111250 - RDC

        :param query: (optional) :class:`basin3d.core.schema.query.QueryMeasurementTimeseriesTVP` object
        :param kwargs: (required) Measurement Timeseries TVP Query parameters. See Query info below.
        :return: a :class:`~basin3d.core.synthesis.DataSourceModelIterator` that yields :class:`basin3d.core.models.MeasurementTimeseriesTVPObservation` objects

        .. note::
            **Measurement Timeseries TVP Query parameters**
            :class:`basin3d.core.schema.query.QueryMeasurementTimeseriesTVP`

            | Required arguments:
            |
            |    * **monitoring_feature** List of monitoring features id(s)
            |    * **observed_property** List of observed property(ies), i.e., BASIN-3D observed property vocabulary. See :func:`basin3d.synthesis.DataSynthesizer.observed_properties`
            |    * **start_date** Start date YYYY-MM-DD
            |
            | Optional arguments:
            |
            |    * **end_date** End date YYYY-MM-DD
            |    * **aggregation_duration** A single aggregation duration :class:`basin3d.core.schema.enum.AggregationDurationEnum` (YEAR|MONTH|DAY|HOUR|MINUTE|SECOND|NONE)
            |    * **statistic** List of statistic(s) :class:`basin3d.core.schema.enum.StatisticEnum` (MEAN|MIN|MAX|INSTANTANEOUS)
            |    * **result_quality** List of result quality(ies) :class:`basin3d.core.schema.enum.ResultQualityEnum` (VALIDATED|UNVALIDATED|SUSPECTED|REJECTED|ESTIMATED)
            |    * **sampling_medium** List of sampling medium(s) :class:`basin3d.core.schema.enum.SamplingMediumEnum` (SOLID_PHASE|WATER|GAS|OTHER)
            |    * **datasource** A single data source id prefix


        .. note::
            **Measurement Timeseries TVP Observation attributes**
            :class:`basin3d.core.models.MeasurementTimeseriesTVPObservation`

            | Attributes values are dependent on data source features.
            |
            | :class:`basin3d.core.models.MappedAttribute` are returned for several attributes so that the data source values are also available.
            |
            |    * **id** Observation identifier
            |    * **type** Type of observation: MEASUREMENT_TVP_TIMESERIES
            |    * **observed_property** The observation's observed property :class:`basin3d.core.models.MappedAttribute`
            |    * **datasource** The data source :class:`basin3d.core.models.DataSource`
            |    * **sampling_medium** Observed property sampling medium :class:`basin3d.core.models.MappedAttribute` (SOLID_PHASE, WATER, GAS, OTHER)
            |    * **phenomenon_time** Datetime of the observation, for a timeseries the start and end times can be provided
            |    * **utc_offset** Coordinate Universal Time offset in hours (offset in hours), e.g., +9
            |    * **feature_of_interest** Monitoring Feature object :class:`basin3d.core.models.MonitoringFeature`, feature on which the observation is being made
            |    * **feature_of_interest_type** Feature type of the feature of interest, :class:`basin3d.core.schema.enum.FeatureTypeEnum`
            |    * **result** Observed values of the observed property being assessed, and (opt) their result quality, :class:`basin3d.core.models.ResultListTVP`
            |    * **time_reference_position** Position of timestamp in aggregated_duration (START, MIDDLE, END)
            |    * **aggregation_duration** Time period represented by observation :class:`basin3d.core.models.MappedAttribute` (YEAR, MONTH, DAY, HOUR, MINUTE, SECOND)
            |    * **unit_of_measurement** Units in which the observation is reported
            |    * **statistic** Statistical property of the observation result :class:`basin3d.core.models.MappedAttribute` (MEAN, MIN, MAX, TOTAL)
            |    * **result_quality** List quality assessment found in the results :class:`basin3d.core.models.MappedAttribute` (VALIDATED, UNVALIDATED, SUSPECTED, REJECTED, ESTIMATED)


        """
        if not query:
            # Raises validation errors
            query = QueryMeasurementTimeseriesTVP(**kwargs)

        #  mypy casts are only used as hints for the type checker,
        #  and they don’t perform a runtime type check.
        return cast(DataSourceModelIterator,
                    self._measurement_timeseries_tvp_observation_access.list(
                        query))



"""

.. currentmodule:: basin3d.core.schema.query

:platform: Unix, Mac
:synopsis: BASIN-3D Query Schema
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top


"""
from datetime import date
from typing import ClassVar, List, Optional, Union

from pydantic import BaseModel, Field

from basin3d.core.schema.enum import FeatureTypeEnum, MessageLevelEnum, ResultQualityEnum, SamplingMediumEnum, StatisticEnum, AggregationDurationEnum


def _to_camelcase(string) -> str:
    """
        Change provided string with underscores to Javascript camelcase
        (e.g. to_camelcase -> toCamelcase)
        :param string: The string to transform
        :return:
        """
    return "".join(i and s[0].upper() + s[1:] or s for i, s in enumerate(string.split("_")))


class QueryBase(BaseModel):
    """ Query Base Class.  This sets `QueryBase.Config` defaults and processes incoming datasource ids"""

    datasource: Optional[List[str]] = Field(title="Datasource Identifiers",
                                            description="List of datasource identifiers to query by.")

    id: Optional[str] = Field(title="Identifier", description="The unique identifier for the desired object")

    is_valid_translated_query: Union[None, bool] = Field(default=None, title="Valid translated query",
                                                         description="Indicates whether the translated query is valid: None = is not translated")

    def __init__(self, **data):
        """
        Custom constructor to modify datasource string to list, if necessary

        :param data: the data
        """
        if "datasource" in data and data['datasource']:
            data['datasource'] = isinstance(data['datasource'], str) and list([data['datasource']]) or data[
                'datasource']
        super().__init__(**data)

    class Config:
        # output fields to camelcase
        alias_generator = _to_camelcase
        # whether an aliased field may be populated by its name as given by the model attribute
        #  (allows bot camelcase and underscore fields)
        allow_population_by_field_name = True
        # Instead of using enum class use enum value (string object)
        use_enum_values = True
        # Validate all fields when initialized
        validate_all = True

    # Get the query fields that have mappings. Subclasses may overwrite this base function
    mapped_fields: ClassVar[List[str]] = []

    # Get the query fields that have prefixes. Subclasses may overwrite ths base function
    prefixed_fields: ClassVar[List[str]] = []


class QueryMonitoringFeature(QueryBase):
    """Query :class:`basin3d.core.models.MonitoringFeature`"""
    # optional but id (QueryBase) is required to query by named monitoring feature
    feature_type: Optional[FeatureTypeEnum] = Field(title="Feature Type",
                                                    description="Filter results by the specified feature type.")
    monitoring_feature: Optional[List[str]] = Field(title="Monitoring Features",
                                                    description="Filter by the list of monitoring feature identifiers")
    parent_feature: Optional[List[str]] = Field(title="Parent Monitoring Features",
                                                description="Filter by the list of parent monitoring feature identifiers")

    def __init__(self, **data):
        """
        Custom constructor to modify feature_type strings to uppercase

        :param data: the data
        """

        # convert strings to lists for some fields; the camel case is for Pydantic validation
        for field in ["monitoring_feature", "monitoringFeature", "parent_feature", "parentFeature"]:
            if field in data and data[field] and isinstance(data[field], str):
                data[field] = list([data[field]])

        # To upper for feature type
        for field in ["featureType", "feature_type"]:
            if field in data and data[field]:
                data[field] = isinstance(data[field], str) and data[field].upper() or data[field]
        super().__init__(**data)

    prefixed_fields: ClassVar[List[str]] = ['id', 'monitoring_feature', 'parent_feature']


class QueryMeasurementTimeseriesTVP(QueryBase):
    """Query :class:`basin3d.core.models.MeasurementTimeseriesTVP`"""
    # required
    monitoring_feature: List[str] = Field(min_items=1, title="Monitoring Features",
                                          description="Filter by the list of monitoring feature identifiers")
    observed_property: List[str] = Field(min_items=1, title="Observed Property Variables",
                                         description="Filter by the list of observed property variables")
    start_date: date = Field(title="Start Date", description="Filter by data taken on or after the start date")

    # optional
    aggregation_duration: AggregationDurationEnum = Field(default='DAY', title="Aggregation Duration",
                                                          description="Filter by the specified aggregation duration or time frequency")
    end_date: Optional[date] = Field(title="End Date", description="Filter by data taken on or before the end date")
    statistic: Optional[List[StatisticEnum]] = Field(title="Statistic",
                                                     description="Return specified statistics, if they exist.")
    result_quality: Optional[List[ResultQualityEnum]] = Field(title="Result Quality",
                                                              description="Filter by specified result qualities")
    sampling_medium: Optional[List[SamplingMediumEnum]] = Field(title="Sampling Medium",
                                                                description="Filter results by specified sampling medium")

    def __init__(self, **data):
        """
        Custom constructor

        :param data: the data
        """

        # convert strings to lists for some fields; the camel case are for Pydantic validation (don't delete)
        for field in ["monitoring_feature", "monitoringFeature", "observed_property", "observedProperty",
                      "statistic", "result_quality", "sampling_medium"]:
            if field in data and data[field] and isinstance(data[field], str):
                data[field] = list([data[field]])

        data = self.__validate__(**data)

        super().__init__(**data)

    @staticmethod
    def __validate__(**data):
        """
        Valiate
        :return:
        """
        if 'aggregation_duration' in data and data['aggregation_duration'] is None:
            del data['aggregation_duration']
        return data

    # observed_property_variables is first b/c it is most likely to have compound mappings.
    # ToDo: check how order may affect translation (see core/synthesis)
    mapped_fields: ClassVar[List[str]] = ['observed_property', 'aggregation_duration', 'statistic', 'result_quality', 'sampling_medium']
    prefixed_fields: ClassVar[List[str]] = ['monitoring_feature']


class SynthesisMessage(BaseModel):
    """BASIN-3D Synthesis Message """

    msg: str = Field(title="Msg", description="The synthesis message ")
    level: MessageLevelEnum = Field(title="Level", description="The severity level of the message.")
    where: Optional[List[str]] = Field([], title="Where",
                                       description="The place in BASIN-3D where the synthesis message was generated "
                                                   "from. "
                                                   "If empty or null, this is a BASIN-3D error, the first item in "
                                                   "the list is the datsource id, "
                                                   "the second should be the synthesis model.")

    class Config:
        # output fields to camelcase
        alias_generator = _to_camelcase
        # whether an aliased field may be populated by its name as given by the model attribute
        #  (allows bot camelcase and underscore fields)
        allow_population_by_field_name = True
        # Instead of using enum class use enum value (string object)
        use_enum_values = True
        # Validate all fields when initialized
        validate_all = True


class SynthesisResponse(BaseModel):
    """BASIN-3D Synthesis Response """

    query: QueryBase = Field(title="Query", description="The original query for the current response")
    data: Optional[Union[object, List[object]]] = Field(title="Data",
                                                        description="The data for the current response. Empty if provided "
                                                                    "via Iterator.")
    messages: List[Optional[SynthesisMessage]] = Field([], title="Messages",
                                                       description="The synthesis messages for this response")

    class Config:
        # output fields to camelcase
        alias_generator = _to_camelcase
        # whether an aliased field may be populated by its name as given by the model attribute
        #  (allows bot camelcase and underscore fields)
        allow_population_by_field_name = True
        # Instead of using enum class use enum value (string object)
        use_enum_values = True
        # Validate all fields when initialized
        validate_all = True
        # Allows generic object to be used for data field
        arbitrary_types_allowed = True
